#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include "system.h"
#include "alt_types.h"
#include "sys/alt_alarm.h"
#include "sys/alt_cache.h"
#include "sys/alt_dev.h"
#include "sys/alt_irq.h"
#include "sys/alt_sys_init.h"
#include "priv/alt_file.h"
#include "altera_avalon_pio_regs.h"
#include "altera_avalon_jtag_uart_regs.h"
#include "altera_avalon_jtag_uart.h"

/*
 * This prototype helps the debugger understand the our alt_main() should be
 * treated like main().
 */
int main (void) __attribute__ ((weak, alias ("alt_main")));

/*
 * Define some values to be used for various timer alarms.
 */
#define QUARTER_SECOND      ( alt_ticks_per_second() / 4 )
#define TENTH_SECOND        ( alt_ticks_per_second() / 10 )
#define TEN_SECOND          ( alt_ticks_per_second() * 10 )
#define JTAG_UART_TIMEOUT   ( alt_ticks_per_second() * 20 )

/*
 * This structure and defines are used by an LED blinker task.
 */
typedef struct {
  volatile alt_u32 led_state;
  volatile alt_u32 led_mask;
} MY_LED_BLINK_STRUCT;

#define LED_MASK_TOGGLE_BIT         (1 << 0)
#define LED_MASK_JTAG_WAITING_BIT   (1 << 1)
#define LED_MASK_JTAG_ABANDONED_BIT (1 << 2)

/*
 * These defines locate our two possible application images at specifically
 * these two offsets in the flash memory.
 */
#define FLASH_IMAGE_1_ADDR  ( EXT_FLASH_BASE + 0x00100000 )
#define FLASH_IMAGE_2_ADDR  ( EXT_FLASH_BASE + 0x00200000 )

/*
 * This example prints a lot of information, this macro is intended to make the
 * code a bit easier to read.
 */
#define JTAGPRINT(name) my_jtag_write(name, sizeof(name)-1)

/*
 * The application images stored in flash memory, have a specific header
 * attached to them.  This is the structure of that header.
 */
typedef struct {
  alt_u32 signature;
  alt_u32 version;
  alt_u32 timestamp;
  alt_u32 data_length;
  alt_u32 data_crc;
  alt_u32 res1;
  alt_u32 res2;
  alt_u32 header_crc;
} MY_FLASH_HEADER;

/*
 * I'm defining a couple globally accessable variables.
 */
volatile MY_LED_BLINK_STRUCT MY_LED_BLINK_CONTEXT;
void *FLASH_IMAGE_1_PTR = FLASH_IMAGE_1_ADDR;
void *FLASH_IMAGE_2_PTR = (void*)FLASH_IMAGE_2_ADDR;

/*
 * These are the function prototypes for all the function references in this
 * example.
 */
extern void my_sys_init( void );

void jump_from_loader(void target(void));
alt_u32 load_flash_image ( int image );
int pick_flash_image(void);
int validate_flash_image(void *image_ptr);
alt_u32 calc_crc32_reflected(alt_u8 *buf, int len);
int my_jtag_write(const char *buf, int len);
int my_jtag_write_8(const char *buf, int len);
alt_u32 my_generic_timeout_callback (void* context);
alt_u32 my_led_blink_callback (void* context);

/*
 * And finally our main() entry point.  We are implementing this as an
 * alt_main() instead of a plain main(), so that we can better control the
 * drivers that load and the system resources that are enabled.  Since this
 * example is booting out of flash ram, and size may be a consideration, this
 * method allows us to keep the memory requirements trim.
 */
int alt_main(void)
{
  /*
   * Define the local variables used by this function.
   */
  alt_alarm my_led_blink_alarm;
  alt_alarm my_ten_second_alarm;
  alt_alarm my_debounce_alarm;
  volatile alt_u32 my_ten_second_context;
  volatile alt_u32 my_debounce_context;
  alt_u32 current_sw0_state;
  alt_u32 cur_button_state;
  int boot_image;
  int waiting_for_debounce;
  alt_u32 entry_point;
  
  /*
   * Define the strings used in this function for printing out STDOUT.
   * 
   * Don't think this ia a normal practice. I only do this to avoid dragging in
   * the printf C library.
   */
  char stars_str[] = "****************************************\n";
  char hello_str0[] = "    Demo Boot Loader Starting - V1.0\n        Booting from Flash\n\n";
  char hello_str1[] = "\nThis loader expects application images\nto be located at offset 1M or 2M in flash\nmemory.\n";
  char hello_str2[] = "\nThis loader will wait 10 seconds before\nattempting to locate a valid application\nimage in flash, to load and execute.\n";
  char hello_str3[] = "\nPressing SW0 forces the loader to\nattempt booting immediately.\n\n";
  char boot_proceed_str[] = "\nNow attempting to boot.\n";
  char reset_str[] = "\nNow jumping back to reset vector.\n";
  char sw0_press_str[] = "\nSW0 button pressed.\n";
  char no_image_str[] = "\nNo valid application image found in\nflash memory.\n";
  char picked_image_str[] = "\nPicked image in flash at location ";
  char image1_str[] = "1M.\n";
  char image2_str[] = "2M.\n";
  char jumping_str[] = "\nNow attempting to load and jump to the\napplication.\n\n";

  /*
   * And now we get some real work done.
   * 
   * Perform system initialization, since we are defining ourselves as an
   * alt_main entry point instead of simply main, it's our responsibility to
   * initialize any drivers we wish to use.  The upside to this, is we can
   * control specifically which drivers we want to initialize.  An example of
   * how to do this can be found in the hello_world freestanding demo.
   */

  /*
   * I'm going to allow interrupts to happen while my boot loader executes, so
   * we must initialize the main irq handler.
   */
  alt_irq_init (ALT_IRQ_BASE);
  
  /*
   * Now I'm going to initialize the drivers that I require, this is stolen
   * directly from the auto generated alt_sys_init.c.
   */
  my_sys_init();
  
  /*
   * I'm going to use some STDIO, so we'll go ahead and set those up to their
   * defaults as selected in the Nios II IDE.
   */
  alt_io_redirect (ALT_STDOUT, ALT_STDIN, ALT_STDERR);
  
  /*
   * Now the system is initialized and ready to use.
   * 
   * At this point you could start doing whatever you feel necessary.  You could
   * perform diagnostics, validate the hardware environment, communicate with
   * higher beings to obtain instructions on how to proceed, etc., etc.
   * 
   * Since this is just a demo, I'm just going to perform a few simple tricks
   * and then boot an application from flash.
   */
  
  /*
   * Setup an led to blink every quarter second, this will use the timer
   * interrupt
   */
  // initialize the MY_LED_BLINK_CONTEXT structure to blink LED0
  MY_LED_BLINK_CONTEXT.led_state = 0x0;
  MY_LED_BLINK_CONTEXT.led_mask = LED_MASK_TOGGLE_BIT;
  // write the initial LED value out to the LEDs
  IOWR_ALTERA_AVALON_PIO_DATA(LED_PIO_BASE, MY_LED_BLINK_CONTEXT.led_state);
  // set an alarm to blink the LEDs every quarter second
  alt_alarm_start ( 
                    &my_led_blink_alarm,            // alt_alarm* alarm,
                    QUARTER_SECOND,                 // alt_u32 nticks,
                    my_led_blink_callback,          // alt_u32 (*callback) (void* context),
                    (void *)&MY_LED_BLINK_CONTEXT   // void* context
                  );

  /*
   * Print a startup message so the user see's we're alive.
   */
  JTAGPRINT(stars_str);
  JTAGPRINT(stars_str);
  JTAGPRINT(hello_str0);
  JTAGPRINT(stars_str);
  JTAGPRINT(hello_str1);
  JTAGPRINT(hello_str2);
  JTAGPRINT(hello_str3);
  JTAGPRINT(stars_str);
  JTAGPRINT(stars_str);
  
  /*
   * We'll wait for 10 seconds, then try to auto boot into an image stored in
   * flash.
   */
  my_ten_second_context = 0x0;
  alt_alarm_start ( 
                    &my_ten_second_alarm,           // alt_alarm* alarm,
                    TEN_SECOND,                     // alt_u32 nticks,
                    my_generic_timeout_callback,    // alt_u32 (*callback) (void* context),
                    (void *)&my_ten_second_context  // void* context
                  );

  /*
   * Wait for the ten second alarm to expire, or for a SW0 pushbutton press.
   */  
  waiting_for_debounce = 0;
  while(my_ten_second_context == 0)
  {
    // read the current state of sw0
    cur_button_state = IORD_ALTERA_AVALON_PIO_DATA(BUTTON_PIO_BASE);
    current_sw0_state = cur_button_state & 0x01;

    if(waiting_for_debounce)  // if we saw sw0 pressed before, then we are waiting_for_debounce
    {
      if(my_debounce_context) // if the debounce alarm has expired
      {
        if(current_sw0_state)
        {
          // the sw0 is currently released, so just clear the debounce wait
          waiting_for_debounce = 0;
        }
        else
        {
          // the sw0 is still pressed, so exit the while loop
          JTAGPRINT(sw0_press_str);
          break;
        }
      }
    }
    else
    {
      // we aren't waiting_for_debounce, if sw0 is currently pressed, then start
      // a debounce alarm and wait for debounce.
      if(!current_sw0_state)
      {
        waiting_for_debounce = 1;
        my_debounce_context = 0x0;
        alt_alarm_start ( 
                          &my_debounce_alarm,           // alt_alarm* alarm,
                          TENTH_SECOND,                 // alt_u32 nticks,
                          my_generic_timeout_callback,  // alt_u32 (*callback) (void* context),
                          (void *)&my_debounce_context  // void* context
                        ); 
      }
    }
  }
  
  /*
   * Either we waited 10 seconds, or SW0 was pressed, either way, we proceed
   * to boot the applicatoin.
   */
  JTAGPRINT(boot_proceed_str);

  if( (boot_image = pick_flash_image()) ) // pick a flash image to load
  {
    // print out what we chose
    JTAGPRINT(picked_image_str);
    if(boot_image == 1)
    {
      JTAGPRINT(image1_str);
    }
    else
    {
      JTAGPRINT(image2_str);
    }

    JTAGPRINT(jumping_str);
    
    /*
     * Now we're going to try to load the application into memory, this will
     * undoubtedly overwrite our current exception handler located in our
     * exception memory.  So before se do that we should probably disable
     * interrupts and not turn them back on again. 
     * 
     * It's also important that we make sure not to rely on things in the
     * ".rwdata" or ".bss" sections, and we should minimize our reliance on the
     * stack as well.  Since all of these sections exist in the exception memory
     * they are all subject to being overwritten.  You should be able to ensure
     * thru inspection of your new exception handler, and the boot loader stack
     * usage, whether or not you will have a problem with this.
     */
    alt_irq_disable_all ();
   
    if( (entry_point = load_flash_image(boot_image)) )  // load the image
    {
      // jump to the entry point of the applicatoin
      jump_from_loader((void(*)(void))(entry_point));
    }
    else
    {
      // if the entry point is not found, then we should just jump back to the
      // reset vector.
      jump_from_loader((void(*)(void))(NIOS2_RESET_ADDR));
    }
  }
  else
  {
    // we didn't find an applicatoin to boot, so just jump back thru the reset
    // vector.
    JTAGPRINT(no_image_str);
    JTAGPRINT(reset_str);
    jump_from_loader((void(*)(void))(NIOS2_RESET_ADDR));
  }
  
  // we should never get here
  exit(0);
}

/*
 * This routine shutsdown the boot loader and jumps to somewhere else.
 */
void jump_from_loader(void target(void))
{
  /*
   * If we had any outstanding I/O or system resources that needed to be cleanly
   * disabled, this would be the place to do that.
   * 
   * In this example we really just need to ensure the state of the Nios II cpu.
   * So if we disable interrupts, and flush the caches, then the program that we
   * jump to should feel like it's receiving the cpu just as it would coming
   * out of a hardware reset.
   */
  alt_irq_disable_all ();
  alt_dcache_flush_all ();
  alt_icache_flush_all ();
  
  /*
   * The cpu state is as clean as we can get it, so now jump to the new
   * applicatoin.
   */
  target();
  
  /*
   * In the odd event that the program we jump to decides to return, we should 
   * probably just jump back thru the reset vector.
   */
  jump_from_loader((void(*)(void))(NIOS2_RESET_ADDR));
}

/*
 * This subroutine loads a flash image from flash, into the Nios II memory map.
 * It basically decodes load records in the format produced from the
 * elf2flash utility, and loads the image as directed by those records.
 * 
 * The input operand, "image", expected to be the image selector indicating
 * which of the 2 possible flash images should be loaded.
 */
alt_u32 load_flash_image ( int image )
{
  alt_u8 *next_flash_byte;
  alt_u32 length;
  alt_u32 address;
  
  // load the image pointer based on the value of "image"
  if( image == 1 )
  {
    next_flash_byte = (alt_u8 *)FLASH_IMAGE_1_PTR + 32;
  }
  else
  {
    next_flash_byte = (alt_u8 *)FLASH_IMAGE_2_PTR + 32;
  }

  /*
   * Now, in this example, I'm guaranteed that my flash images will be located
   * at word alignment within the flash memory, but that's not necessarily going
   * to be the case in other implementations.  So instead of writing this
   * routine to take advantage of that assumption, I'm going to write it
   * assuming that I could be dealing with byte alignment at best.
   * 
   * I'll use memcpy to transfer the bytes from flash to whereever they're
   * going.
   */
   
  // get the first 4 bytes, which should be a length record
  memcpy( (void*)(&length), (void*)(next_flash_byte), (size_t)(4) );
  next_flash_byte += 4;
  
  // now loop until I get an entry record, or a halt recotd
  while( (length != 0) && (length != 0xffffffff) )
  {
    // get the next 4 bytes, which should be a address record
    memcpy( (void*)(&address), (void*)(next_flash_byte), (size_t)(4) );
    next_flash_byte += 4;
    
    // copy the next "length" bytes to "address"
    memcpy( (void*)(address), (void*)(next_flash_byte), (size_t)(length) );
    next_flash_byte += length;
    
    // get the next 4 bytes, which should be a length record
    memcpy( (void*)(&length), (void*)(next_flash_byte), (size_t)(4) );
    next_flash_byte += 4;
  }
  
  // we didn't read a length record, soooo
  if( length == 0xffffffff )
  {
    // we read a HALT record, so return a 0
    return 0;
  }
  else
  {
    // we got an entry record, so read the next 4 bytes for the entry address
    memcpy( (void*)(&address), (void*)(next_flash_byte), (size_t)(4) );
    next_flash_byte += 4;
    
    // and return the entry address value
    return address;
  }
}

/*
 * This subroutine scans the two possible flash images and picks one to be
 * loaded by the boot loader.  The selection criteria are pretty simple:
 *    1 - if there's only one valid image, then we choose that one.
 *    2 - if there are two valid images then
 *          A - We choose the one with the latest revision
 *          B - if both images have the same revision, the we choose the one
 *              with the latest time stamp.
 *          C - if both time stamp values are equal, the we choose image #2
 *    3 - if neither image is valid, we return 0
 */
int pick_flash_image(void)
{
  MY_FLASH_HEADER temp_header_1 __attribute__((aligned(4)));
  MY_FLASH_HEADER temp_header_2 __attribute__((aligned(4)));
  
  // start by validating image 1
  if( validate_flash_image((alt_u32 *)(FLASH_IMAGE_1_PTR)) )
  {
    // image 1 was not valid, so validate image 2
    if( validate_flash_image((alt_u32 *)(FLASH_IMAGE_2_PTR)) )
    {
      // image 2 is not valid, return 0
      return 0;
    }
    else
    {
      // image 2 is valid and image 1 was not, so choose 2
      return 2;
    }
  }
  else
  {
    // image 1 was valid, check image 2
    if( validate_flash_image((alt_u32 *)(FLASH_IMAGE_2_PTR)) )
    {
      // image 2 was not valid, so choose image 1
      return 1;
    }
    else
    {
      /*
       * both images were valid, so we now must compare header information.
       * Just like the load_flash subroutine above, we don't assume that the
       * flash images are word aligned, so we memcpy the headers from flash, to
       * our word aligned temporary header buffers.
       */
      memcpy(&temp_header_1, FLASH_IMAGE_1_PTR, 32);
      memcpy(&temp_header_2, FLASH_IMAGE_2_PTR, 32);
      
      // first compare the versions
      if( temp_header_1.version == temp_header_2.version )
      {
        // the versions are equal, so compare timestamps
        if( temp_header_1.timestamp > temp_header_2.timestamp )
        {
          // image 1 has a later timestamp
          return 1;
        }
        else
        {
          // image 2 has a later timestamp, or the timestamps were equal
          return 2;
        }
      }
      else
      {
        // the versions were not equal
        if( temp_header_1.version > temp_header_2.version )
        {
          // image 1 has a later version
          return 1;
        }
        else
        {
          // image 2 has a later version
          return 2;
        }
      }
    }
  }
}

/*
 * This routine validates flash images.
 */
int validate_flash_image(void *image_ptr)
{
  MY_FLASH_HEADER temp_header  __attribute__((aligned(4)));
  
  /*
   * Again, we don't assume the image is word aligned, so we copy the header
   * from flash to a word aligned buffer.
   */
  memcpy(&temp_header, image_ptr, 32);
  
  // check the signature first
  if( temp_header.signature == 0xa5a5a5a5 )
  {
    // signature is good, validate the header crc
    if( calc_crc32_reflected( (alt_u8*)&temp_header, 32) )
    {
      // header crc does not validate, return 2
      return 2;
    }
    else
    {
      // header crc is valid, now validate the data crc
      if ( temp_header.data_crc == calc_crc32_reflected( image_ptr + 32, temp_header.data_length) )
      {
        // data crc validates, the image is good
        return 0;
      }
      else
      {
        // data crc is not valid, return 3
        return 3;
      }
    }
  }
  else
  {
    // bad signature, return 1
    return 1;
  }
}

/*
 * This subroutine calcuates a reflected CRC32 on the data in the buffer passed
 * into it.
 */
alt_u32 calc_crc32_reflected(alt_u8 *buf, int len)
{
  alt_u32 crcval = 0xffffffff;
  int i;
  alt_u8 cval;
  
  while(len--)
  { 
    cval = *buf++;
    crcval ^= cval;
    for (i = 8; i--; )
      crcval = (crcval & 0x00000001) ? ((crcval >> 1) ^ 0xEDB88320) : (crcval >> 1);
  }
  return crcval;
}

/*
 * This subroutine is provided so that we can write characters out the JTAG UART
 * without getting into a situation where we are permanently blocked, because
 * there is no host reading from the JTAG UART.  So this routine takes the
 * message that we want to write, and breaks it up into 8 byte or less chunks
 * to send to my_jtag_write_8.  my_jtag_write_8 makes sure that we can write
 * to the jtag uart before attempting the write.
 */
int my_jtag_write(const char *buf, int len)
{
  int ret_val;
  int orig_len = len;
  
  if(len > 8)
  {
    // our write message is greater than 8 bytes long, so break it up into
    // 8-byte or less chunks.
    do
    {
      ret_val = my_jtag_write_8( buf, (len > 8)?(8):(len));
      if(ret_val < 0)
      {
        // if an error returns then we're done
        break;
      }
      else
      {
        // no error, decrement our length, and increment the buffer
        len -= ret_val;
        buf += ret_val;
      }
    } while(len > 0);
    
    // return the whole length of the buffer transmitted
    ret_val = orig_len;
    
  }
  else
  {
    // the write message is less than 8 bytes, just write it.
    ret_val = my_jtag_write_8( buf, len);
  }
  
  return ret_val;
}

/*
 * This subroutine is called by my_jtag_write, which feeds it 8-byte or less
 * chunks of data to write out the JTAG UART.  Basically, this routine queries
 * the write fifo in the jtag uart to make sure there is room for our data
 * first.  If there's not room, then it waits for up to 20 seconds to see if a
 * host will clear some of the pending data to allow us to write new data to the
 * fifo.  If no host connects within 20 seconds, then the jtag uart is marked
 * abandoned, and the write data is discarded.  Once abandoned, every time a new
 * write call is made to this routine, it checks to see if a host may have
 * emptied the write fifo, if that ever happens, then the jtag uart is reclaimed
 * and the original algorithm prevails.
 */
int my_jtag_write_8(const char *buf, int len)
{
  alt_fd *the_fd;
  alt_avalon_jtag_uart_dev *the_dev;
  unsigned int the_base;
  alt_u32 control;
  alt_u32 wr_fifo_space;
  int ret_val;
  alt_alarm my_jtag_uart_alarm;
  volatile alt_u32 my_jtag_uart_context;
  
  /*
   * fish thru the device table to find our device block and extract the base
   * address to this STDOUT peripheral.
   * 
   * A big assumption here is that the STDOUT peripheral is a JTAG UART.
   */
  the_fd = &alt_fd_list[STDOUT_FILENO];
  the_dev = (alt_avalon_jtag_uart_dev *)the_fd->dev;
  the_base = the_dev->base;

  // read the jtag uart control register and grab the write fifo space
  control = IORD_ALTERA_AVALON_JTAG_UART_CONTROL(the_base);
  wr_fifo_space = (control & ALTERA_AVALON_JTAG_UART_CONTROL_WSPACE_MSK)
                  >> ALTERA_AVALON_JTAG_UART_CONTROL_WSPACE_OFST;

  if(wr_fifo_space >= len)
  {
    // the write fifo has room for our write, so clear the abandoned flag
    // and write the message.
    MY_LED_BLINK_CONTEXT.led_mask &= ~LED_MASK_JTAG_ABANDONED_BIT;
    
    ret_val = write(STDOUT_FILENO, buf, len);
  } 
  else
  {
    // the write fifo does not have room for us. have we previously abandoned
    // the jtag uart?
    if( MY_LED_BLINK_CONTEXT.led_mask & LED_MASK_JTAG_ABANDONED_BIT )
    {
      // we have previously abandoned the jtag uart
      if(control & ALTERA_AVALON_JTAG_UART_CONTROL_AC_MSK)
      {
        // there has been activity from a host so let's clear the abandoned
        // flag, and see if the host will clear the fifo for us.
        MY_LED_BLINK_CONTEXT.led_mask &= ~LED_MASK_JTAG_ABANDONED_BIT;
      }
    }
    
    // at this point we check to see if the jtag uart is abandoned
    if( MY_LED_BLINK_CONTEXT.led_mask & LED_MASK_JTAG_ABANDONED_BIT )
    {
      // the jtag uart has been abandoned, so just dump the data and return -1
      ret_val = -1;
    }
    else
    {
      // the jtag uart has not been abandoned
      ret_val = -1;
      
      // clear the activity bit in the jtag uart control register
      control |= ALTERA_AVALON_JTAG_UART_CONTROL_AC_MSK;
      IOWR_ALTERA_AVALON_JTAG_UART_CONTROL(the_base, control);
      
      // set the jtag waiting bit, so LED1 blinks for the user
      MY_LED_BLINK_CONTEXT.led_mask |= LED_MASK_JTAG_WAITING_BIT;
      
      // set a 20 second timeout alarm
      my_jtag_uart_context = 0;
      alt_alarm_start ( 
                        &my_jtag_uart_alarm,          // alt_alarm* alarm,
                        JTAG_UART_TIMEOUT,            // alt_u32 nticks,
                        my_generic_timeout_callback,  // alt_u32 (*callback) (void* context),
                        (void *)&my_jtag_uart_context // void* context
                      );
      
      // now wait until the timeout occurs and abandon the uart, or the host
      // clears the fifo for us.
      while( my_jtag_uart_context == 0 )
      {
        // get the current control register value
        control = IORD_ALTERA_AVALON_JTAG_UART_CONTROL(the_base);
        
        if( control & ALTERA_AVALON_JTAG_UART_CONTROL_AC_MSK )
        {
          // we see activity, so stop the timeour alarm
          alt_alarm_stop ( &my_jtag_uart_alarm );
          my_jtag_uart_context = 0;
          
          // extract the write fifo space
          wr_fifo_space = (control & ALTERA_AVALON_JTAG_UART_CONTROL_WSPACE_MSK)
                          >> ALTERA_AVALON_JTAG_UART_CONTROL_WSPACE_OFST;
          
          if(wr_fifo_space >= len)
          {
            // we now have room to perform our write, so do it and get outa here
            ret_val = write(STDOUT_FILENO, buf, len);
            break;
          } 
          // there's still not enough room so clear the activity bit
          control |= ALTERA_AVALON_JTAG_UART_CONTROL_AC_MSK;
          IOWR_ALTERA_AVALON_JTAG_UART_CONTROL(the_base, control);
          
          // and set the timeout alarm again
          alt_alarm_start ( 
                            &my_jtag_uart_alarm,          // alt_alarm* alarm,
                            JTAG_UART_TIMEOUT,            // alt_u32 nticks,
                            my_generic_timeout_callback,  // alt_u32 (*callback) (void* context),
                            (void *)&my_jtag_uart_context // void* context
                          );
          
        }
      }
      
      // well if we get here, then we're done waiting, so clear the waiting flag
      MY_LED_BLINK_CONTEXT.led_mask &= ~LED_MASK_JTAG_WAITING_BIT;
  
      if( my_jtag_uart_context == 1 )
      {
        // we got here because the timeout alarm fired, so set the abandoned flag
        MY_LED_BLINK_CONTEXT.led_mask |= LED_MASK_JTAG_ABANDONED_BIT;
      }
    }
  }
  return ret_val;  
}

/*
 * This subroutine is my generic timeout callback routine for timeout alarms
 * that get set.  This routine simply increments the alt_u32 pointed to by the
 * context pointer and returns 0, which requests no additional alarm time.
 */
alt_u32 my_generic_timeout_callback (void* context)
{
  *((volatile alt_u32 *)(context)) += 0x1;
  return(0);
}

/*
 * This is the alarm callback routine for my led blinker alarm.  All this does
 * is toggle the leds and return, requesting another quarter second alarm.
 */
alt_u32 my_led_blink_callback (void* context)
{

  ((MY_LED_BLINK_STRUCT *)(context))->led_state ^= ((MY_LED_BLINK_STRUCT *)(context))->led_mask;

  if( ((MY_LED_BLINK_STRUCT *)(context))->led_state & LED_MASK_TOGGLE_BIT )
  {
    ((MY_LED_BLINK_STRUCT *)(context))->led_state &= ((MY_LED_BLINK_STRUCT *)(context))->led_mask;
  }
  else
  {
    ((MY_LED_BLINK_STRUCT *)(context))->led_state = 0x0;
  }
  
  IOWR_ALTERA_AVALON_PIO_DATA(LED_PIO_BASE, ((MY_LED_BLINK_STRUCT *)(context))->led_state);

  return(QUARTER_SECOND);
}

