#include <stdio.h>
#include "system.h"
#include "alt_types.h"
#include "sys/alt_alarm.h"
#include "altera_avalon_pio_regs.h"
#include "sys/alt_cache.h"
#include "sys/alt_irq.h"
#include "sys/alt_flash.h"
#include "string.h"
#include "time.h"

/*
 * Define some version values.
 */
#define MAJOR_VERSION 1
#define MINOR_VERSION 0

/*
 * This structure and defines are used by an LED blinker task.
 */
typedef struct {
  volatile alt_u32 led_state;
  volatile alt_u32 led_mask;
} MY_LED_BLINK_STRUCT;

#define LED_MASK_TOGGLE_BIT         (1 << 7)

/*
 * These defines locate our two possible application images at specifically
 * these two offsets in the flash memory.
 */
#define FLASH_IMAGE_1_ADDR  ( EXT_FLASH_BASE + 0x00100000 )
#define FLASH_IMAGE_2_ADDR  ( EXT_FLASH_BASE + 0x00200000 )

/*
 * This is the structure of flash header that gets appended to the applications
 * before they are programmed into flash.
 */
typedef struct {
  alt_u32 signature;
  alt_u32 version;
  alt_u32 timestamp;
  alt_u32 data_length;
  alt_u32 data_crc;
  alt_u32 res1;
  alt_u32 res2;
  alt_u32 header_crc;
} MY_FLASH_HEADER;

/*
 * Define some values to be used for various timer alarms.
 */
#define QUARTER_SECOND      ( alt_ticks_per_second() / 4 )
#define TENTH_SECOND        ( alt_ticks_per_second() / 10 )
#define TWO_SECOND          ( alt_ticks_per_second() * 2 )
#define THIRTY_SECOND       ( alt_ticks_per_second() * 30 )

/*
 * These are the prototypes for the functions in this program.
 */
void jump_to_loader(void);
void print_new_image_info(void);
int try_downloading(void);
void program_flash(MY_FLASH_HEADER *target_flash_image, MY_FLASH_HEADER *source_image);
int pick_flash_image(void);
int validate_flash_image(void *image_ptr);
alt_u32 calc_crc32_reflected(alt_u8 *buf, int len);
alt_u32 my_generic_timeout_callback (void* context);
alt_u32 my_led_blink_callback (void* context);

/*
 * I'm defining a couple globally accessable variables.
 */
void *FLASH_IMAGE_1_PTR = (void*)FLASH_IMAGE_1_ADDR;
void *FLASH_IMAGE_2_PTR = (void*)FLASH_IMAGE_2_ADDR;
unsigned char sdram_inbuffer[1024*1024] __attribute__((aligned(4)));

int main()
{
  volatile MY_LED_BLINK_STRUCT MY_LED_BLINK_CONTEXT;
  alt_alarm my_led_blink_alarm;
  alt_alarm my_thirty_second_alarm;
  alt_alarm my_debounce_alarm;
  volatile alt_u32 my_thirty_second_context;
  volatile alt_u32 my_debounce_context;
  alt_u32 current_sw3_state;
  alt_u32 cur_button_state;
  int waiting_for_debounce;
  int target_flash;
  int check_flash;
  
  /*
   * Setup an led to blink every quarter second, this will use the timer
   * interrupt
   */
  // initialize the MY_LED_BLINK_CONTEXT structure to blink LED0
  MY_LED_BLINK_CONTEXT.led_state = 0x0;
  MY_LED_BLINK_CONTEXT.led_mask = LED_MASK_TOGGLE_BIT;
  // write the initial LED value out to the LEDs
  IOWR_ALTERA_AVALON_PIO_DATA(LED_PIO_BASE, MY_LED_BLINK_CONTEXT.led_state);
  // set an alarm to blink the LEDs every quarter second
  alt_alarm_start ( 
                    &my_led_blink_alarm,            // alt_alarm* alarm,
                    QUARTER_SECOND,                 // alt_u32 nticks,
                    my_led_blink_callback,          // alt_u32 (*callback) (void* context),
                    (void *)&MY_LED_BLINK_CONTEXT   // void* context
                  );

  /*
   * Print some status information so the user knows we're alive.
   */
  printf("++++++++++++++++++++++++++++++++++++++++\n");
  printf("++++++++++++++++++++++++++++++++++++++++\n");
  printf("    Loaded Application - V%02d.%02d\n", MAJOR_VERSION, MINOR_VERSION);
  printf("    Running from SDRAM\n");
  printf("++++++++++++++++++++++++++++++++++++++++\n");
  printf("\nThis application waits for 30 seconds\n");
  printf("and then jumps back to the boot loader.\n");
  printf("\nWhile the application is waiting, you\n");
  printf("can download a flash image to the\n");
  printf("application over the download JTAG UART.\n");
  printf("If this application receives a valid\n");
  printf("flash image over the download port, then\n");
  printf("it will program that image into flash\n");
  printf("memory and jump back to the boot loader.\n");
  printf("You can use the following command to\n");
  printf("write a flash image file to the download\n");
  printf("JTAG UART:\n");
  printf("\n[SOPC Builder]$ cat -B <your_flash_image> | nios2-terminal --cable=1 --device=1 --instance=0\n");
  printf("\nIf you press SW3 at any time, the\n");
  printf("application will jump back to the boot\n");
  printf("loader.\n");
  printf("++++++++++++++++++++++++++++++++++++++++\n");
  printf("++++++++++++++++++++++++++++++++++++++++\n");

  /*
   * We'll wait for 30 seconds, then try jumping back to the boot loader.
   */
  my_thirty_second_context = 0x0;
  alt_alarm_start ( 
                    &my_thirty_second_alarm,           // alt_alarm* alarm,
                    THIRTY_SECOND,                     // alt_u32 nticks,
                    my_generic_timeout_callback,       // alt_u32 (*callback) (void* context),
                    (void *)&my_thirty_second_context  // void* context
                  );

  /*
   * Wait for the ten second alarm to expire, or for a SW3 pushbutton press.
   */  
  waiting_for_debounce = 0;
  while(my_thirty_second_context == 0)
  {
    // each time through the loop we see if there's any download data coming
    // in the download jtag uart.
    if( !( try_downloading() ) )
    {
      // we got a download so, let's print the structure of the new image
      print_new_image_info();
      
      // let's pick a location in flash to stick it.
      target_flash = pick_flash_image();
      
      printf("\nNow programming flash image at 0x%08lX\n", ( (target_flash == 0) || (target_flash == 2) )?((alt_u32)FLASH_IMAGE_1_PTR):((alt_u32)FLASH_IMAGE_2_PTR));
      
      // program the image into the appropriate location
      if( (target_flash == 0) || (target_flash == 2) )
      {
        program_flash((MY_FLASH_HEADER *)FLASH_IMAGE_1_PTR, (MY_FLASH_HEADER *)sdram_inbuffer);
        // validate the image after we program it.
        check_flash = validate_flash_image((alt_u32 *)(FLASH_IMAGE_1_PTR));
      }
      else
      {
        program_flash((MY_FLASH_HEADER *)FLASH_IMAGE_2_PTR, (MY_FLASH_HEADER *)sdram_inbuffer);
        // validate the image after we program it.
        check_flash = validate_flash_image((alt_u32 *)(FLASH_IMAGE_2_PTR));
      }
      
      if(check_flash)
      {
        printf("\nNew flash image does NOT validate.\n");
      }
      else
      {
        printf("\nNew flash image validates.\n");
      }
      
      // now exit the while loop so we can jump back to the boot loader
      break;
    }
    
    // read the current state of sw3
    cur_button_state = IORD_ALTERA_AVALON_PIO_DATA(BUTTON_PIO_BASE);
    current_sw3_state = cur_button_state & 0x08;

    if(waiting_for_debounce)  // if we saw sw3 pressed before, then we are waiting_for_debounce
    {
      if(my_debounce_context) // if the debounce alarm has expired
      {
        if(current_sw3_state)
        {
          // the sw3 is currently released, so just clear the debounce wait
          waiting_for_debounce = 0;
        }
        else
        {
          // the sw3 is still pressed, so exit the while loop
          printf("\nSW3 Pressed\n");
          break;
        }
      }
    }
    else
    {
      // we aren't waiting_for_debounce, if sw3 is currently pressed, then start
      // a debounce alarm and wait for debounce.
      if(!current_sw3_state)
      {
        waiting_for_debounce = 1;
        my_debounce_context = 0x0;
        alt_alarm_start ( 
                          &my_debounce_alarm,           // alt_alarm* alarm,
                          TENTH_SECOND,                 // alt_u32 nticks,
                          my_generic_timeout_callback,  // alt_u32 (*callback) (void* context),
                          (void *)&my_debounce_context  // void* context
                        ); 
      }
    }
  }
  
  // we're done with the work so lets get back to the boot loader
  printf("\nApplication is finished waiting.\n");
  printf("\nNow returning to boot loader.\n\n");
  
  jump_to_loader();
  
  return 0;
}

/*
 * This routine shutsdown the application and jumps to the boot loader
 */
void jump_to_loader(void)
{
  alt_alarm my_two_second_alarm;
  volatile alt_u32 my_two_second_context;
  
  /*
   * If we had any outstanding I/O or system resources that needed to be cleanly
   * disabled, this would be the place to do that.
   * 
   * I'm going to wait 2 seconds to allow any write data to flush from the I/O
   * drivers.
   */
  my_two_second_context = 0x0;
  alt_alarm_start ( 
                    &my_two_second_alarm,           // alt_alarm* alarm,
                    TWO_SECOND,                     // alt_u32 nticks,
                    my_generic_timeout_callback,       // alt_u32 (*callback) (void* context),
                    (void *)&my_two_second_context  // void* context
                  );
                  
  while(my_two_second_context == 0);
  
  /*
   * In this example we really just need to ensure the state of the Nios II cpu.
   * So if we disable interrupts, and flush the caches, then the program that we
   * jump to should feel like it's receiving the cpu just as it would coming
   * out of a hardware reset.
   */
  alt_irq_disable_all ();
  alt_dcache_flush_all ();
  alt_icache_flush_all ();
  
  /*
   * The cpu state is as clean as we can get it, so now jump to the boot loader
   */
  ((void(*)(void))NIOS2_RESET_ADDR)();
}

/*
 * 
 */
void print_new_image_info(void)
{
  MY_FLASH_HEADER *header = (MY_FLASH_HEADER*)sdram_inbuffer;
  alt_u8 *buf = sdram_inbuffer;
  alt_u32 length;
  alt_u32 address;
  
  if( validate_flash_image((void *)sdram_inbuffer) )
  {
    printf("\nCannot print the download info, image does NOT validate.\n");
  }
  else
  {
    // print the header info
    printf("\n\
    New Flash Image Info:\n\
\n\
      signature = 0x%08lX\n\
        version = 0x%08lX\n\
      timestamp = 0x%08lX\n\
    data_length = 0x%08lX\n\
       data_crc = 0x%08lX\n\
           res1 = 0x%08lX\n\
           res2 = 0x%08lX\n\
     header_crc = 0x%08lX\n\
\n\
    The timestamp value = UTC %s",
          header->signature,
          header->version,
          header->timestamp,
          header->data_length,
          header->data_crc,
          header->res1,
          header->res2,
          header->header_crc,
          ctime(&header->timestamp)
    );
  
    // increment the buf pointer past the header
    buf = buf +32;
    
    // get the first 4 bytes, which should be a length record
    memcpy( (void*)(&length), (void*)(buf), (size_t)(4) );
    buf += 4;
    
    // now loop until I get an entry record, or a halt recotd
    while( (length != 0) && (length != 0xffffffff) )
    {
      // get the next 4 bytes, which should be a address record
      memcpy( (void*)(&address), (void*)(buf), (size_t)(4) );
      buf += 4;
      
      printf("\n  -- Data Record --\n        LOAD ADDR:  0x%08lX\n           LENGTH:  0x%08lX -- %ld_d\n", address, length, length);

      // increment buf past the data section
      buf += length;
      
      // get the next 4 bytes, which should be a length record
      memcpy( (void*)(&length), (void*)(buf), (size_t)(4) );
      buf += 4;
    }
    
    // we didn't read a length record, soooo
    if( length == 0xffffffff )
    {
      // we read a HALT record
      printf("\n  -- HALT Record --\n\n");
    }
    else
    {
      // we got an entry record, so read the next 4 bytes for the entry address
      memcpy( (void*)(&address), (void*)(buf), (size_t)(4) );
      buf += 4;
      printf("\n  -- Entry Record --\n       ENTRY ADDR:  0x%08lX\n\n", address);
    }
  }

}

/* 
 * This routine is called to try downloading a flash image.
 */
int try_downloading(void)
{
  FILE *download_fp;
  int cur_read_len;
  int read_len;
  alt_u32 data_calc_crc32;
  char *read_buf = (void *)sdram_inbuffer;
  MY_FLASH_HEADER *header_buf = (MY_FLASH_HEADER *)sdram_inbuffer;
  int receiving_flag;

  /*
   * attempt to open the download jtag uart
   */
  download_fp = fopen(DOWNLOAD_JTAG_UART_NAME, "r+");
  if((download_fp == NULL))
  {
    printf("ERROR: Failed to open the download JTAG UART \"%s\"\n", DOWNLOAD_JTAG_UART_NAME);
    return (1);
  }
  
  /*
   * attempt to read the header bytes into SDRAM
   */
  receiving_flag = 1;
  read_len = 32;
  while(read_len)
  {
    cur_read_len = fread(read_buf, 1, read_len, download_fp);
    if(cur_read_len > 0)
    {
      // we are getting some data, print a message once to inform the user
      if(receiving_flag)
      {
        printf("\nReceiving data on download JTAG UART.\n");
        receiving_flag = 0;
      }
      read_len -= cur_read_len;
      read_buf += cur_read_len;
    }
    else
    {
      // when we don't get any data, we just return with an error value
      if(read_len == 32)
      {
        fclose(download_fp);
        return (1);
      }
    }
  }

  printf("\nHeader received.\n");
  
  // after we get the header downloaded, we validate it.
  if(header_buf->signature != 0xa5a5a5a5)
  {
    printf("\nABORTING DOWNLOAD: Bad Header Signature Detected.\n");
    fprintf(download_fp, "\n\nDownload failed.\n\n\x04"); // terminate nios2-terminal
    fclose(download_fp);
    return (1);
  }
  
  printf("\nHeader signature is valid.\n");
  
  if(calc_crc32_reflected((alt_u8 *)(header_buf), 32))
  {
    printf("\nABORTING DOWNLOAD: Bad Header CRC Detected.\n");
    fprintf(download_fp, "\n\nDownload failed.\n\n\x04"); // terminate nios2-terminal
    fclose(download_fp);
    return (1);
  }

  printf("\nHeader CRC is valid.\n");

  // attempt to read the rest of the binary data
  printf("\nReading data block of %d bytes.\n", (int)header_buf->data_length);
  read_len = header_buf->data_length;
  while(read_len)
  {
    cur_read_len = fread( read_buf, 1, read_len, download_fp);
    if(cur_read_len > 0)
    {
      read_len -= cur_read_len;
      read_buf += cur_read_len;
    }
  }
  
  printf("\nData block received.\n");
  
  // validate the data block CRC
  data_calc_crc32 = calc_crc32_reflected((((alt_u8 *)(header_buf)) + 32), header_buf->data_length);
  if(data_calc_crc32 != header_buf->data_crc)
  {
    printf("\nABORTING DOWNLOAD: Bad Data CRC Detected.\n");
    fprintf(download_fp, "\n\nDownload failed.\n\n\x04"); // terminate nios2-terminal
    fclose(download_fp);
    return (1);
  }
  
  printf("\nData CRC is valid.\n");

  fprintf(download_fp, "\n\nDownload succeeded.\n\n\x04"); // terminate nios2-terminal
  fclose(download_fp);
 
  printf("\nFlash image successfully downloaded.\n");
 
  return (0);
}

/*
 * This routine programs a new image into flash.
 */
void program_flash(MY_FLASH_HEADER *target_flash_image, MY_FLASH_HEADER *source_image)
{
  alt_flash_fd *flash_fd;
  int flash_offset;
  
  flash_offset = (alt_u32)target_flash_image - (alt_u32)EXT_FLASH_BASE;
  
  flash_fd = alt_flash_open_dev(EXT_FLASH_NAME);
  if(flash_fd)
  {
    alt_write_flash(  
                      flash_fd,                              // alt_flash_fd* fd,
                      flash_offset,                          // int offset,
                      (void*)source_image,                   // const void* src_addr,
                      (int)(source_image->data_length + 32)  // int length
                   );
    
    alt_flash_close_dev(flash_fd);
  }
  else
  {
    printf("\nERROR: could not open the flash driver.\n");
  }
}

/*
 * This subroutine scans the two possible flash images and picks one to be
 * loaded by the boot loader.  The selection criteria are pretty simple:
 *    1 - if there's only one valid image, then we choose that one.
 *    2 - if there are two valid images then
 *          A - We choose the one with the latest revision
 *          B - if both images have the same revision, the we choose the one
 *              with the latest time stamp.
 *          C - if both time stamp values are equal, the we choose image #2
 *    3 - if neither image is valid, we return 0
 */
int pick_flash_image(void)
{
  MY_FLASH_HEADER temp_header_1 __attribute__((aligned(4)));
  MY_FLASH_HEADER temp_header_2 __attribute__((aligned(4)));
  
  // start by validating image 1
  if( validate_flash_image((alt_u32 *)(FLASH_IMAGE_1_PTR)) )
  {
    // image 1 was not valid, so validate image 2
    if( validate_flash_image((alt_u32 *)(FLASH_IMAGE_2_PTR)) )
    {
      // image 2 is not valid, return 0
      return 0;
    }
    else
    {
      // image 2 is valid and image 1 was not, so choose 2
      return 2;
    }
  }
  else
  {
    // image 1 was valid, check image 2
    if( validate_flash_image((alt_u32 *)(FLASH_IMAGE_2_PTR)) )
    {
      // image 2 was not valid, so choose image 1
      return 1;
    }
    else
    {
      /*
       * both images were valid, so we now must compare header information.
       * Just like the load_flash subroutine above, we don't assume that the
       * flash images are word aligned, so we memcpy the headers from flash, to
       * our word aligned temporary header buffers.
       */
      memcpy(&temp_header_1, FLASH_IMAGE_1_PTR, 32);
      memcpy(&temp_header_2, FLASH_IMAGE_2_PTR, 32);
      
      // first compare the versions
      if( temp_header_1.version == temp_header_2.version )
      {
        // the versions are equal, so compare timestamps
        if( temp_header_1.timestamp > temp_header_2.timestamp )
        {
          // image 1 has a later timestamp
          return 1;
        }
        else
        {
          // image 2 has a later timestamp, or the timestamps were equal
          return 2;
        }
      }
      else
      {
        // the versions were not equal
        if( temp_header_1.version > temp_header_2.version )
        {
          // image 1 has a later version
          return 1;
        }
        else
        {
          // image 2 has a later version
          return 2;
        }
      }
    }
  }
}

/*
 * This routine validates flash images.
 */
int validate_flash_image(void *image_ptr)
{
  MY_FLASH_HEADER temp_header  __attribute__((aligned(4)));
  
  /*
   * Again, we don't assume the image is word aligned, so we copy the header
   * from flash to a word aligned buffer.
   */
  memcpy(&temp_header, image_ptr, 32);
  
  // check the signature first
  if( temp_header.signature == 0xa5a5a5a5 )
  {
    // signature is good, validate the header crc
    if( calc_crc32_reflected( (alt_u8*)&temp_header, 32) )
    {
      // header crc does not validate, return 2
      return 2;
    }
    else
    {
      // header crc is valid, now validate the data crc
      if ( temp_header.data_crc == calc_crc32_reflected( image_ptr + 32, temp_header.data_length) )
      {
        // data crc validates, the image is good
        return 0;
      }
      else
      {
        // data crc is not valid, return 3
        return 3;
      }
    }
  }
  else
  {
    // bad signature, return 1
    return 1;
  }
}

/*
 * This subroutine calcuates a reflected CRC32 on the data in the buffer passed
 * into it.
 */
alt_u32 calc_crc32_reflected(alt_u8 *buf, int len)
{
  alt_u32 crcval = 0xffffffff;
  int i;
  alt_u8 cval;
  
  while(len--)
  { 
    cval = *buf++;
    crcval ^= cval;
    for (i = 8; i--; )
      crcval = (crcval & 0x00000001) ? ((crcval >> 1) ^ 0xEDB88320) : (crcval >> 1);
  }
  return crcval;
}

/*
 * This subroutine is my generic timeout callback routine for timeout alarms
 * that get set.  This routine simply increments the alt_u32 pointed to by the
 * context pointer and returns 0, which requests no additional alarm time.
 */
alt_u32 my_generic_timeout_callback (void* context)
{
  *((volatile alt_u32 *)(context)) += 0x1;
  return(0);
}

/*
 * This is the alarm callback routine for my led blinker alarm.  All this does
 * is toggle the leds and return, requesting another quarter second alarm.
 */
alt_u32 my_led_blink_callback (void* context)
{

  ((MY_LED_BLINK_STRUCT *)(context))->led_state ^= ((MY_LED_BLINK_STRUCT *)(context))->led_mask;

  if( ((MY_LED_BLINK_STRUCT *)(context))->led_state & LED_MASK_TOGGLE_BIT )
  {
    ((MY_LED_BLINK_STRUCT *)(context))->led_state &= ((MY_LED_BLINK_STRUCT *)(context))->led_mask;
  }
  else
  {
    ((MY_LED_BLINK_STRUCT *)(context))->led_state = 0x0;
  }
  
  IOWR_ALTERA_AVALON_PIO_DATA(LED_PIO_BASE, ((MY_LED_BLINK_STRUCT *)(context))->led_state);

  return(QUARTER_SECOND);
}
